module: state


// Constants.
define constant <path-cost> = <integer>;

////////////////////////////////////////////////////////////
//
// The following defines a <point>.
define constant <coordinate> = limited(<integer>, min: 0);


define functional sealed class <point> (<object>)
  slot val :: <integer>,
    required-init-keyword: value:;
end class <point>;

define sealed domain make(singleton(<point>));
define sealed domain initialize(<point>);


define sealed class <board> (<object>)
  slot board-width :: <integer>,
    required-init-keyword: width:;
  slot board-height :: <integer>,
    required-init-keyword: height:;
  slot board-bits :: <byte-string> = make(<byte-string>, size: 1024*1024);
end;


define sealed class <robot> (<object>)
  slot robot-id :: <integer>,
    required-init-keyword: id:;
  slot robot-pos :: false-or(<point>),
    init-value: #f, init-keyword: pos:;
end;


define sealed class <package> (<object>)
  slot pkg-id :: <integer>,
    required-init-keyword: id:;
  slot location :: <point>,
    required-init-keyword: location:;
  
  // things we don't know unless we've seen it
  slot dest :: false-or(<point>),
    init-value: #f, init-keyword: dest:;
  slot weight :: false-or(<integer>),
    init-value: #f, init-keyword: weight:;
  slot carrier :: false-or(<integer>), 
    init-value: #f, init-keyword: carrier:;
end class <package>;



define sealed class <state> (<object>)
  slot board :: <board>,
    required-init-keyword: board:;
  slot me :: <robot>,
    required-init-keyword: me:;
  slot robot-money :: <integer>,
    init-value: 0, init-keyword: money:;
  slot my-capacity :: <integer>,
    required-init-keyword: my-capacity:;
  slot my-packages :: <list>,
    init-value: #(), init-keyword: my-packages:;

  slot robots :: <list>,
    init-value: #(), init-keyword: robots:;
  slot bases :: <list>,
    init-value: #(), init-keyword: bases:;
  slot packages :: <list>,
    init-value: #(), init-keyword: packages:;
end;

define sealed domain make(singleton(<board>));
define sealed domain make(singleton(<robot>));
define sealed domain make(singleton(<package>));
define sealed domain make(singleton(<state>));

define sealed domain initialize(<board>);
define sealed domain initialize(<robot>);
define sealed domain initialize(<package>);
define sealed domain initialize(<state>);


define method print-object(r :: <robot>, stream :: <stream>)
 => ();
  format(stream, "<robot #%= %=>",
         r.robot-id,
         r.robot-pos);
end;

define method print-object(b :: <board>, stream :: <stream>)
 => ();
  format(stream, "<board #%= x %=\n", b.board-width, b.board-height);
  for (y from b.board-height + 1 to 0 by -1)
    for (x from 0 to b.board-width + 1)
      let pt = new-point(x, y);
      format(stream, "|%c%d|",
             cell-char(b, pt),
             cell-cost(b, pt));
    end;
    format(stream, "\n");
  end;
  format(stream, ">\n");
end;

define method print-object(s :: <state>, stream :: <stream>)
 => ();
  format(stream, "<state me:%= $%= cap:%=, #pack:%=, #base:%=, #pack:%=",
         s.me,
         s.robot-money,
         s.my-capacity,
         s.my-packages.size,
         s.bases.size,
         s.packages.size);
  format(stream, "\nPackages we have: ");
  for (p in s.my-packages)
    format(stream, "\n  %= ", p);
  end;

  format(stream, "\nPackages we know about: ");
  for (p in s.packages)
    format(stream, "\n  %= ", p);
  end;

  for (r in s.robots)
    format(stream, "\n  %=", r)
  end;
  format(stream, "\n");
end;


///////////////////////////////////////////////////////
//
// Board stuff.
//
// bits 0-1: empty = 00  base  = 10
//           wall  = 01  water = 11
// bits 2-3: cost: 0 = normal  1 = next to a wall 3 = next to water

define inline function passable?(board :: <board>, i :: <point>)
 => res :: <boolean>;
  without-bounds-checks
    logand(as(<integer>, board.board-bits[i.val]), 1) = 0;
  end;
end;

define inline function cell-cost(board :: <board>, i :: <point>)
 => res :: <integer>;
  without-bounds-checks
    ash(as(<integer>, board.board-bits[i.val]), -2) + 10;
  end
end;

define method cell-char(board :: <board>, i :: <point>)
 => ch :: <character>;
  select(logand(as(<integer>, board.board-bits[i.val]), 3))
    0 => '.';
    1 => '#';
    2 => '@';
    3 => '~';
  end;
end;


define method initialize
    (b :: <board>, #next next-method, #key width, height)
 => ();
  next-method();
  dd("in board initialize, width=%= height=%=\n", width, height);
  for (x from 1 to width)
    for (y from 1 to height)
      b.board-bits[new-point(x, y).val] := as(<character>, 0);
    end;
  end;
end;


define method set-cell
    (board :: <board>, p :: <point>, ch :: <character>)
 => ();
  let code =
    select(ch)
      '.' => 0;
      '#' => 1;
      '@' => 2;
      '~' => 3;
    end;
  board.board-bits[p.val] :=
    as(<character>, logand(as(<integer>, board.board-bits[p.val]), #xfc) + code);

  let (x, y) = coords(p);
  if (logand(code, 1) = 1)
    for (i from -1 to 1)
      for (j from -1 to 1)
        let (px, py) = values(x + i, y + j);
        if (px > 0 & py > 0)
          let pt = new-point(px, py);
          board.board-bits[pt.val] := 
            as(<character>, as(<integer>, board.board-bits[pt.val]) + ash(code,2));
        end;
      end;
    end;
  end;
end;

define class <not-supplied> (<object>) end;
define constant $not-supplied = make(<not-supplied>);


define inline function update-robot
    (it :: <robot>, #key
       new-pos = $not-supplied)
 => res :: <robot>;
  make(<robot>,
       id: it.robot-id,
       pos: if (new-pos == $not-supplied) it.robot-pos else new-pos end)
end;

define inline function find-robot(s :: <state>, id :: <integer>)
 => res :: false-or(<robot>);
  any?(method(it :: <robot>) it.robot-id = id & it end, s.robots);
end;

define inline function find-package(s :: <state>, id :: <integer>)
 => res :: false-or(<package>);
  any?(method(it :: <package>) it.pkg-id = id & it end, s.packages);
end;

define inline function update-state
    (it :: <state>, #key
       new-my-packages = $not-supplied,
       new-bases = $not-supplied,
       new-money = $not-supplied,
       new-packages = $not-supplied,
       new-robots = $not-supplied)
 => res :: <state>;

  // make sure we keep me updated
  let new-me = it.me;
  if (new-robots ~= $not-supplied)
    block (done)
      for (r in new-robots)
        if (r.robot-id == it.me.robot-id)
          new-me := r;
          done();
        end;
      end;
    end;
  end;

  make(<state>,
       board: it.board,
       me: new-me,
       my-capacity: it.my-capacity,
       money: if (new-money == $not-supplied) it.robot-money else new-money end,
       my-packages: if (new-my-packages == $not-supplied)
                      it.my-packages else new-my-packages end,
       bases: if (new-bases == $not-supplied) it.bases else new-bases end,
       robots: if (new-robots == $not-supplied) it.robots else new-robots end,
       packages: if (new-packages == $not-supplied) it.packages else new-packages end)
end;


define inline function my-location(s :: <state>)
 => res :: <point>;
  s.me.robot-pos | new-point(1,1);
end;


////////////////////////////////////////////////////////////

// Useful distance-cost function for the board. Manhattan cost to avoid square root.
define inline function distance-cost(source :: <point>,
                                     target :: <point>)
 => (res :: <path-cost>)
  abs(source.x - target.x) + abs(source.y - target.y);
end function distance-cost;


define inline function x(p :: <point>) => (x :: <coordinate>);
  logand(p.val, #x3ff);
end;


define inline function y(p :: <point>) => (y :: <coordinate>);
  logand(ash(p.val, -10), #x3ff);
end;


define inline function new-point(x :: <coordinate>, y :: <coordinate>)
 => (res :: <point>);
  make(<point>, value: logior(ash(y, 10), logand(x, #x3ff)));
end;

define inline function coords(pt :: <point>)
 => (x :: <integer>, y :: <integer>);
  values(pt.x, pt.y);
end;

define inline function move-west(pt :: <point>)
 => res :: <point>;
  make(<point>, value: pt.val - 1);
end;

define inline function move-east(pt :: <point>)
 => res :: <point>;
  make(<point>, value: pt.val + 1);
end;

define inline function move-north(pt :: <point>)
 => res :: <point>;
  make(<point>, value: pt.val + 1024);
end;

define inline function move-south(pt :: <point>)
 => res :: <point>;
  make(<point>, value: pt.val - 1024);
end;


define inline sealed method \=(a :: <point>, b :: <point>)
 => b :: <boolean>;
  a.val == b.val;
end method;


define method print-object(p :: <point>, stream :: <stream>)
 => ();
  format(stream, "(%d,%d)", p.x, p.y);
end;

///////////////////////////////////////////

define inline method at-destination?(p :: <package>) 
 => res :: <boolean>;
  p.dest & p.dest = p.location;
end method at-destination?;

define inline sealed method update-package(it :: <package>, 
	#key new-id = $not-supplied,
	     new-weight = $not-supplied,
	     new-location = $not-supplied,
	     new-dest = $not-supplied,
	     new-carrier = $not-supplied)
 => (it :: <package>);
  make(<package>,
       id:        it.pkg-id,
       weight:    it.weight,
       location:  if (new-location == $not-supplied) it.location else new-location end,
       dest:      it.dest,
       carrier:   if (new-carrier  == $not-supplied) it.carrier else new-carrier end);
end method update-package;


define method print-object (p :: <package>, s :: <stream>) => ()
  format(s, "{<package> id: %d= weight: %=, location: %=, dest: %=, carrier: %=}",
         p.pkg-id, p.weight, p.location, p.dest, p.carrier);
end method print-object;


define function packages-at-location(s :: <state>, where :: <point>)
  => packages :: <list>;
  choose(method (p :: <package>) p.location = where end, s.packages);
end;

define function locations-having-packages(s :: <state>)
  => points :: <list>;
  map(method(a :: <package>) a.location end,
      remove-duplicates!
        (sort(s.packages,
              test: method(a :: <package>, b :: <package>)
                        a.location.val < b.location.val
                    end),
         test: method(a :: <package>, b :: <package>)
                   a.location.val = b.location.val
               end));
end;

